unit HotKeyManager;

{ Some methods have moved to the Classes unit in D6 and are thus deprecated.
  Using the following compiler directives we handle that situation. }
{$IFDEF VER140}
  {$DEFINE DELPHI_6}
  {$DEFINE DELPHI_6_UP}
{$ENDIF}

interface

uses
  Classes, Windows, Messages;

type
  TOnHotKeyPressed = procedure(HotKey: TShortCut; Index: Word) of object;

  PHotKeyRegistration = ^THotKeyRegistration;
  THotKeyRegistration = record
    HotKey: TShortCut;
    KeyIndex: Word;
  end;

  THotKeyManager = class(TComponent)
  private
    FHandle: HWND;
    HotKeyList: TList;
    FOnHotKeyPressed: TOnHotKeyPressed;
    OldAppProc, NewAppProc: Pointer;   // Procedure variables
    // Hook methods
    procedure HookApp;
    procedure UnhookApp;
    procedure HookAppProc(var Msg: TMessage);
    function DisposeHotKey(hkr: PHotKeyRegistration): Boolean;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function AddHotKey(HotKey: TShortCut): Word;
    function ChangeHotKey(Index: Word; NewHotKey: TShortCut): Word;
    function RemoveHotKey(HotKey: TShortCut): Boolean;
    function RemoveHotKeyByIndex(Index: Word): Boolean;
    procedure ClearHotKeys;
  published
    property OnHotKeyPressed: TOnHotKeyPressed read FOnHotKeyPressed write FOnHotKeyPressed;
  end;

function HotKeyAvailable(HotKey: TShortCut): Boolean;
//function HotKeyValid(HotKey: TShortCut): Boolean;
procedure SeparateHotKey(HotKey: Word; var Modifiers, Key: Word);

procedure Register;


implementation

uses
  Forms, SysUtils;

const
  HotKeyAtomPrefix = 'HotKeyManagerHotKey';

procedure SeparateHotKey(HotKey: Word; var Modifiers, Key: Word);
// Separate key and modifiers, so they can be used with RegisterHotKey
const
  VK2_SHIFT   =  32;
  VK2_CONTROL =  64;
  VK2_ALT     = 128;
  VK2_WIN     = 256;
var
  Virtuals: Integer;
  V: Word;
  x: Byte;
begin
  Key := Byte(HotKey);
  x := HotKey shr 8;
  Virtuals := x;
  V := 0;
  if (Virtuals and VK2_ALT) <> 0 then
    V := V + MOD_ALT;
  if (Virtuals and VK2_CONTROL) <> 0 then
    V := V + MOD_CONTROL;
  if (Virtuals and VK2_SHIFT) <> 0 then
    V := V + MOD_SHIFT;
  if (Virtuals and VK2_WIN) <> 0 then
    V := V + MOD_WIN;
  Modifiers := V;
end;


function HotKeyAvailable(HotKey: TShortCut): Boolean;
// Test if HotKey is available (test if it can be registered - by this or any app.)
var
  V1, V2: Word;
  Atom: Word;
begin
  Atom := GlobalAddAtom(PChar('HotKeyManagerHotKeyTest'));
  SeparateHotKey(HotKey, V1, V2);
  Result := RegisterHotKey(Application.Handle, Atom, V1, V2);
  if Result then
    UnregisterHotKey(Application.Handle, Atom);
  GlobalDeleteAtom(Atom);
end;

{
function HotKeyValid(HotKey: TShortCut): Boolean;
// Test if HotKey is valid (test if it can be registered even if app. already registered it)
var
  V1, V2: Word;
  WasRegistered: Boolean;
  Atom: Word;
begin
  Atom := GlobalAddAtom(PChar(HotKeyAtomPrefix + IntToStr(HotKey)));
  SeparateHotKey(HotKey, V1, V2);
  WasRegistered := UnregisterHotKey(Application.Handle, Atom);
  if WasRegistered then
  begin
    RegisterHotKey(Application.Handle, Atom, V1, V2);
    Result := True;
  end
  else
  begin
    Result := RegisterHotKey(Application.Handle, Atom, V1, V2);
    if Result then
      UnregisterHotKey(Application.Handle, Atom);
  end;
  GlobalDeleteAtom(Atom);
end;
}
{------------------- THotKeyManager -------------------}

constructor THotKeyManager.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FHandle := Application.Handle;
  HotKeyList := TList.Create;
  // Set hook
  if not (csDesigning in ComponentState) then
    HookApp;                 // Hook into the app.'s message handling
end;


destructor THotKeyManager.Destroy;
begin
  ClearHotKeys;
  HotKeyList.Free;
  // It is important to unhook any hooked processes
  if not (csDesigning in ComponentState) then
    UnhookApp;
  inherited Destroy;
end;


function THotKeyManager.AddHotKey(HotKey: TShortCut): Word;
var
  hkr: PHotKeyRegistration;
  Modifiers, Key: Word;
  Atom: Word;
begin
  SeparateHotKey(HotKey, Modifiers, Key);
  // Create unique id (global atom)
  Atom := GlobalAddAtom(PChar(HotKeyAtomPrefix + IntToStr(HotKey)));
  // Register
  if RegisterHotKey(FHandle, Atom, Modifiers, Key) then
  begin
    hkr := New(PHotKeyRegistration);
    hkr.HotKey := HotKey;
    hkr.KeyIndex := Atom;
    HotKeyList.Add(hkr);
    Result := Atom;
  end
  else
  begin
    GlobalDeleteAtom(Atom);
    Result := 0;
  end;
end;


function THotKeyManager.ChangeHotKey(Index: Word; NewHotKey: TShortCut): Word;
var
  I: Integer;
  hkr: PHotKeyRegistration;
begin
  Result := 0;
  for I := 0 to HotKeyList.Count -1 do
  begin
    hkr := PHotKeyRegistration(HotKeyList[I]);
    if hkr.KeyIndex = Index then
    begin
      RemoveHotKeyByIndex(hkr.KeyIndex);
      Result := AddHotKey(NewHotKey);
      Exit;
    end;
  end;
end;


function THotKeyManager.RemoveHotKey(HotKey: TShortCut): Boolean;
var
  I: Integer;
  hkr: PHotKeyRegistration;
begin
  Result := False;
  for I := 0 to HotKeyList.Count -1 do
  begin
    hkr := PHotKeyRegistration(HotKeyList[I]);
    if hkr.HotKey = HotKey then
    begin
      Result := DisposeHotKey(hkr);
      HotKeyList.Remove(hkr);
      Exit;
    end;
  end;
end;


function THotKeyManager.RemoveHotKeyByIndex(Index: Word): Boolean;
var
  I: Integer;
  hkr: PHotKeyRegistration;
begin
  Result := False;
  for I := 0 to HotKeyList.Count -1 do
  begin
    hkr := PHotKeyRegistration(HotKeyList[I]);
    if hkr.KeyIndex = Index then
    begin
      Result := DisposeHotKey(hkr);
      HotKeyList.Remove(hkr);
      Exit;
    end;
  end;
end;


procedure THotKeyManager.ClearHotKeys;
var
  I: Integer;
  hkr: PHotKeyRegistration;
begin
  for I := HotKeyList.Count -1 downto 0 do
  begin
    hkr := PHotKeyRegistration(HotKeyList[I]);
    DisposeHotKey(hkr);
    HotKeyList.Remove(hkr);
  end;
end;


function THotKeyManager.DisposeHotKey(hkr: PHotKeyRegistration): Boolean;
begin
  // Unregister using previously assigned id (global atom)
  Result := UnregisterHotKey(FHandle, hkr.KeyIndex);
  GlobalDeleteAtom(hkr.KeyIndex);
  if Result then
    Dispose(hkr);
end;


procedure THotKeyManager.HookApp;
begin
  // Hook the application
  OldAppProc := Pointer(GetWindowLong(Application.Handle, GWL_WNDPROC));
{$IFDEF DELPHI_6_UP}
  NewAppProc := Classes.MakeObjectInstance(HookAppProc);
{$ELSE}
  NewAppProc := MakeObjectInstance(HookAppProc);
{$ENDIF}
  SetWindowLong(Application.Handle, GWL_WNDPROC, LongInt(NewAppProc));
end;


procedure THotKeyManager.UnhookApp;
begin
  if Assigned(OldAppProc) then
    SetWindowLong(Application.Handle, GWL_WNDPROC, LongInt(OldAppProc));
  if Assigned(NewAppProc) then
{$IFDEF DELPHI_6_UP}
    Classes.FreeObjectInstance(NewAppProc);
{$ELSE}
    FreeObjectInstance(NewAppProc);
{$ENDIF}
  NewAppProc := nil;
  OldAppProc := nil;
end;


{ All app. messages pass through HookAppProc. You can override the messages
  by not passing them along to Windows (via CallWindowProc). }

procedure THotKeyManager.HookAppProc(var Msg: TMessage);

  function HotKeyFound(HotKey: TShortCut): Boolean;
  var
    I: Integer;
  begin
    Result := False;
    for I := 0 to HotKeyList.Count -1 do
      if PHotKeyRegistration(HotKeyList[I]).HotKey = HotKey then
      begin
        Result := True;
        Break;
      end;
  end;

var
  Modifier: Word;
begin
  case Msg.Msg of
    WM_HOTKEY:
      if Assigned(FOnHotKeyPressed) then
      begin
        // Get modifier keys status
        Modifier := 0;
        if (Msg.LParamLo and MOD_SHIFT) <> 0 then
          Modifier := Modifier + scShift;
        if (Msg.LParamLo and MOD_CONTROL) <> 0 then
          Modifier := Modifier + scCtrl;
        if (Msg.LParamLo and MOD_ALT) <> 0 then
          Modifier := Modifier + scAlt;
        { Check if the hotkey is in the list (it's possible user has added hotkeys
          without using this component and handles these hotkeys by himself). }
        if HotKeyFound(Msg.LParamHi + Modifier) then
          OnHotKeyPressed(Msg.LParamHi + Modifier, Msg.WParam);
      end;
  end;

  // Pass the message on
  Msg.Result := CallWindowProc(OldAppProc, Application.Handle,
                Msg.Msg, Msg.WParam, Msg.LParam);
end;


procedure Register;
begin
  RegisterComponents('System', [THotKeyManager]);
end;

end.

